/*****************************************************************************

NAME:
   bogofilter.c -- detect spam and bogons presented on standard input.

AUTHOR:
   Eric S. Raymond <esr@thyrsus.com>

THEORY:
   This is Paul Graham's variant of Bayes filtering described at 

	http://www.paulgraham.com/spam.html

I do the lexical analysis slightly differently, however.

******************************************************************************/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "bogofilter.h"

// like the stock Judy error macro, except with exit(2)
#define JUDYERROR(CallerFile, CallerLine, JudyFunc, JudyErrno, JudyErrID) \
	{							\
	    (void) fprintf(stderr, "File '%s', line %d: %s(), "	\
			   "bogofilter: JU_ERRNO_* == %d, ID == %d\n",	\
			   CallerFile, CallerLine,		\
			   JudyFunc, JudyErrno, JudyErrID);	\
	    exit(2);						\
	}


#include <Judy.h>

// implementation details
#define HEADER		"# bogofilter wordlist (format version A): %lu\n"

// constants for the Graham formula 
#define GOOD_BIAS	2		// give good words more weight
#define KEEPERS		15		// how many extrema to keep
#define MINIMUM_FREQ	5		// minimum freq
#define UNKNOWN_WORD	0.2f		// odds that unknown word is spammish
#define SPAM_CUTOFF	0.9f		// if it's spammier than this...

#define DEVIATION(n)	fabs((n) - 0.5f)	// deviation from average

#define max(x, y)	(((x) > (y)) ? (x) : (y))
#define min(x, y)	(((x) < (y)) ? (x) : (y))

wordlist_t good_list	= {"good"};
wordlist_t bad_list	= {"bad"};

#define	PLURAL(count) ((count == 1) ? "" : "s")

static dump(wordlist_t *words, FILE *fpout)
{
    char	tokenbuffer[BUFSIZ];
    void	**loc;

    for (loc  = JudySLFirst(words->list, tokenbuffer, 0);
	 loc != (void *) NULL;
	 loc  = JudySLNext(words->list, tokenbuffer, 0))
    {
	(void) fprintf(fpout, "%s %lu\n", tokenbuffer, (*((PWord_t) loc)));
    }

}

static void strlwr(char* s)
{
	char c;
	while((c = *s) != 0)
		*s++ = tolower(c);
}

static void increment(char *word,  Word_t incr, wordlist_t *list)
/* increment a word usage count in the specified list */
{
    void  ** PPValue;			// associated with Index.
    void  *  PArray = (Pvoid_t) NULL;	// JudySL array.
    JError_t JError;                    // Judy error structure

    if ((PPValue = JudySLIns(&(list->list), word, &JError)) == PPJERR)
    {					// assume out of memory.
	(void) printf ("Error %d, cannot insert \"%s\" in array.\n",
		       JU_ERRNO(&JError), word);
	exit(2);
    }

    (*((PWord_t) PPValue)) += incr;	// increment usage count.

    if (verbose >= 3)
	(void) printf ("\"%s\" stored %ld time%s\n",
		   word,  *((PWord_t) PPValue),
		   PLURAL (*((PWord_t) PPValue)));
}

static int getcount(char *word, wordlist_t *list)
/* get the count associated with a given word in a list */
{
    Word_t * PValue;                     // JudySL array element

    JSLG(PValue, list->list, word);
    if (PValue == NULL)
    {
	if (verbose >= 3)
	    printf("getcount: no %s hits for %s\n", list->name, word);
	return 0;
    }
    else
    {
	if (verbose >= 2)
	    printf("getcount: '%s' has %d %s hits in %d\n", 
		   word, *PValue, list->name, list->msgcount);
	return *PValue;
    }
}

void read_list(char *file, wordlist_t *list)
/* initialize one of the associative arrays from a file */
{
    char	tokenbuffer[BUFSIZ];
    FILE	*infp;
    Word_t	count;

    list->file = strdup(file);
    infp = fopen(file, "r");
    if (infp == NULL)
    {
	perror("bogofilter (read_list)");
	exit(2);
    }

    fscanf(infp, HEADER, &list->msgcount);
    while (fscanf(infp, "%s %lu", tokenbuffer, &count) == 2)
	increment(tokenbuffer, count, list);

    fclose(infp);
}

void write_list(wordlist_t *list)
/* dump the state of one of the arrays to a specified file */
{
    char	tokenbuffer[BUFSIZ];
    void	**loc;
    FILE	*outfp;
    char	tempfile[BUFSIZ];

    strncpy(tempfile, list->file, BUFSIZ-10);
    strcat(tempfile, "__temp");
    outfp = fopen(tempfile, "w");
    if (outfp == NULL)
    {
	perror("bogofilter (write_list)");
	exit(2);
    }

    (void) fprintf(outfp, HEADER, list->msgcount);
    dump(list, outfp);
    fclose(outfp);
    unlink(list->file);
    if (rename(tempfile, list->file)) {
	perror("bogofilter (write_list rename)");
        exit(2);
    }
}

void register_words(int fd, wordlist_t *list)
/* tokenize text on stdin and register it to  a specified list */
{
    int	tok;

    yyin = fdopen(fd, "r");
    while ((tok = get_token()) != 0)
    {
	if (tok == FROM)
	    list->msgcount++;

	strlwr(yytext);
	increment(yytext, 1, list);
    }

    if (verbose >= 3)
	printf("register_words: %d messages entered in %s\n", 
	       list->msgcount, list->name);
}

int bogofilter(int fd)
/* evaluate text for spamicity */
{
    double prob, product, invproduct;
    int keepercount = 0;
    double goodness, badness;
    
    int tok;

    typedef struct 
    {
	char	*key;
	double	prob;
    }
    discrim_t;
    static discrim_t extrema[KEEPERS];
    discrim_t *pp;

    for (pp = extrema; pp < extrema+sizeof(extrema)/sizeof(*extrema); pp++)
    {
	pp->prob = 0.5f;
	pp->key = NULL;
    }

    yyin = fdopen(fd, "r");
    while ((tok = get_token()) != 0)
    {
	double dev;

	strlwr(yytext);
	goodness = getcount(yytext, &good_list);
	badness  = getcount(yytext, &bad_list);

	// Paul Graham's formula:
	// 
	// (let ((g (* 2 (or (gethash word good) 0))) 
	//      (b (or (gethash word bad) 0)))
	//  (unless (&lt; (+ g b) 5) 
	//   (max .01 (min .99 
	//  	    (double (/ 
	// 		    (min 1 (/ b nbad)) 
	// 		    (+ (min 1 (/ g ngood)) (min 1 (/ b nbad)))))))))
	goodness *= GOOD_BIAS;
	if (goodness + badness < MINIMUM_FREQ)
	    prob = UNKNOWN_WORD;
	else
	{
	    register double pb = min(1, (badness / bad_list.msgcount));
	    register double pg = min(1, (goodness / good_list.msgcount));

	    prob = pb / (pg + pb);
	    prob = min(prob, 0.99);
	    prob = max(prob, 0.01);
	}

	// update the list of tokens with maximum deviation
	dev = DEVIATION(prob);
	for (pp = extrema; pp < extrema+sizeof(extrema)/sizeof(*extrema); pp++)
	    if (dev > DEVIATION(pp->prob))
	    {
		pp->prob = prob;
		if (verbose)
		{
		    if (pp->key)
			free(pp->key);
		    pp->key = strdup(yytext);
		}
		break;
	    }
    }

    if (verbose)
    {
	for (pp = extrema; pp < extrema+sizeof(extrema)/sizeof(*extrema); pp++)
	    if (pp->key)
		printf("%s -> %f\n", pp->key, pp->prob);
    }

    // Bayes' theorem
    product = invproduct = 1;
    for (pp = extrema; pp < extrema+sizeof(extrema)/sizeof(*extrema); pp++)
	if (pp->prob == 0)
	    break;
    	else
	{
	    product *= pp->prob;
	    invproduct *= (1 - pp->prob);
	}
    prob = product / (product + invproduct);

    if (verbose)
	printf("Spamicity of %f\n", prob);

    return((prob > SPAM_CUTOFF) ? 1 : 0);
}

/* Done */
